/**
 * @file
 *
 * @copyright Axiros GmbH 2018-2021, All rights reserved
 *
 * @brief Main USP handler (aka engine) function definitions
 */

#ifndef USP_ENGINE_H_
#define USP_ENGINE_H_

#include <stdint.h>
#include <time.h>

#include "compiler.h"

#include "mtp.h"


/* Forward declarations */
struct frontend;
struct dm_object;
struct usp_agent;
struct usp_mtp;
struct Usp__Msg;


/**
 *  @brief Our private usp_engine structure
 */
struct usp_engine;


/**
 * Definition of a MTP state change callback function to be called whenever a
 * certain state change occurs */
typedef void (*mtp_state_change_cb) (struct usp_mtp *mtp, void *data);


/**
 * @brief Prepares a new USP engine for running
 *
 * @param[in] fe	the frontend to attach this engine to
 *
 * @return a new USP engine object
 */
struct usp_engine *usp_engine_new (struct frontend *fe) ARGS_NONNULL RETURNS_NONNULL;


/**
 * @brief Return the frontend this engine is attached to
 *
 * @param[in] engine    the engine to get the frontend for
 *
 * @return the frontend which this engine is attached to
 */
struct frontend *usp_engine_get_frontend (struct usp_engine *engine) ARGS_NONNULL RETURNS_NONNULL;


/**
 * @brief Return the data model queue assigned to this engine
 *
 * @param[in] engine    the engine to get the assigned data model queue for
 *
 * @return the data model queue assigned to this engine
 */
struct dm_queue *usp_engine_get_dm_queue (struct usp_engine *engine) ARGS_NONNULL RETURNS_NONNULL;


/**
 * @brief Return the endpoint id associated with this engine via the agent
 *
 * @param[in] engine    the engine to get the associated endpoint id for
 *
 * @return a pointer to the endpoint id associated with this engine via the agent or NULL if no agent is assigned
 */
char *usp_engine_get_endpoint_id (struct usp_engine *engine) ARGS_NONNULL;


/**
 * @brief Add a USP agent to this engine
 *
 * @param[in] engine    the engine to add the new agent to
 * @param[in] agent	the USP agent to add to the specified engine
 */
void usp_engine_add_agent (struct usp_engine *engine, struct usp_agent *agent) ARGS_NONNULL;


/**
 * @brief Add a new MTP to this engine
 *
 * @param[in] engine    the engine to add the new MTP to
 * @param[in] epid	the USP Controller Endpoint ID associated with this MTP
 * @param[in] mtp	the USP MTP to add to the specified engine
 */
void usp_engine_add_mtp (struct usp_engine *engine, char *epid, struct usp_mtp *mtp) ARGS_NONNULL;


/**
 * @brief Remove an MTP from the registry
 *
 * @param[in] engine    the engine to remove the MTP from
 * @param[in] mtp	the USP MTP to remove from the specified engine
 */
void usp_engine_remove_mtp (struct usp_engine *engine, const struct usp_mtp *mtp) ARGS_NONNULL;


/**
 * @brief Define a callback to be executed on MTP connection state change
 *
 * @param[in] engine    the engine on which to run the callback
 * @param[in] callback	the callback function to execute when the state changes
 * @param[in] data	an opaque passthrough data pointer
 */
void usp_engine_add_mtp_state_change_cb (struct usp_engine *engine, mtp_state_change_cb callback, void *data) ARGINDEX_NONNULL(1) ARGINDEX_NONNULL(2);


/**
 * @brief Callback for an MTP to notify engine about a connection state change
 *
 * @param[in] engine    the engine to which the MTP is attadhed to
 * @param[in] mtp	the USP MTP for which the connection state has changed
 * @param[in] state	the new MTP connection state
 */
void usp_engine_notify_mtp_state_change (struct usp_engine *engine, struct usp_mtp *mtp, enum usp_mtp_connection_state state) ARGS_NONNULL;


/**
 * @brief Enable writing of all incoming and outgoing USP Records to filesystem.
 *
 * @note Do NOT use this in production! This function is additionally protected by the DEBUGGING setting to prevent accidental use.
 *
 * @param[in] engine	the engine to enable the record logging on
 */
void usp_engine_enable_record_logs (struct usp_engine *engine) ARGS_NONNULL;


/**
 * @brief Handle an incoming USP record
 *
 * @param[in] engine	the engine used to work on the record
 * @param[in] buf	the received USP Record data
 * @param[in] buflen	the length in bytes of the received data
 */
void usp_engine_handle_record (struct usp_engine *engine, char *buf, size_t buflen) ARGS_NONNULL;


/**
 * @brief Remove an record for which we received a reply (either response or error) from the queue
 *
 * @param[in] engine	the engine to find the associated message for
 * @param[in] msg_id	the message id to look out for
 */
void usp_engine_handle_reply (struct usp_engine *engine, char *msg_id) ARGS_NONNULL;


/**
 * @brief Queue outgoing USP message
 *
 * @param[in] engine		engine to queue the record on
 * @param[in] from		the sender endpoint id
 * @param[in] to		the receiver endpoint id
 * @param[in] msg		the USP message to queue
 * @param[in] send_resp		do we expect a response to the message?
 * @param[in] time_offset	delay in s to defer the sending of a message, use 0 for right now
 */
void usp_engine_queue_msg (struct usp_engine *engine, char *from, char *to, struct Usp__Msg *msg, int send_resp, int time_offset, time_t retry_end) ARGS_NONNULL;


/**
 * @brief Return a handle to the outgoing USP message queue
 *
 * @param[in] engine	the engine to return the queue for
 */
struct usp_queue *usp_engine_get_queue (struct usp_engine *engine) ARGS_NONNULL RETURNS_NONNULL;


/**
 * @brief Return a MTP configured for a particular endpoint id
 *
 * @param[in] engine	the engine to search the configured MTP in
 * @param[in] epid	the endpoint id to match
 *
 * @return the found MTP or NULL if not matched
 */
struct usp_mtp *usp_engine_get_mtp_by_epid (struct usp_engine *engine, char *epid) ARGS_NONNULL;


/**
 * @brief Return the remote endpoint ID a particular MTP is used for
 *
 * @param[in] engine	the engine to search the configured MTP in
 * @param[in] mtp	the MTP to match
 *
 * @return the found endpoint ID or NULL if not matched
 */
char *usp_engine_get_epid_by_mtp (struct usp_engine *engine, const struct usp_mtp *mtp) ARGS_NONNULL;


/**
 * @brief Process the current USP queue
 *
 * @param[in] engine	the engine to process the queue for
 */
void usp_engine_process_queue (struct usp_engine *engine, time_t now) ARGS_NONNULL;


/**
 * @brief Process all pending records for an MTP which just came online
 *
 * @param[in] engine	the engine to process the queue for
 * @param[in] mtp	the MTP which just came online
 */
void usp_engine_flush_records (struct usp_engine *engine, const struct usp_mtp *mtp) ARGS_NONNULL;


/**
 * @brief Sync up changes in the datamodel with the engine bi-directionally
 *
 * @param[in] engine	the engine to sync up
 */
void usp_engine_datamodel_sync (struct usp_engine *engine) ARGS_NONNULL;


/**
 * @brief Automatically configure MTPs according to datamodel information
 *
 * @param[in] engine	the engine to automatically configure
 * @param[in] ws_log	the filename of the WebSocket log file to use or NULL if none desired
 */
void usp_engine_mtp_autoconfig (struct usp_engine *engine, char *ws_log) ARGINDEX_NONNULL(1);


/**
 * @brief Process a successful async command execution and send notifications as required
 *
 * @param[in] engine	the engine used to process the command execution
 * @param[in] reqobj	the associated request object related to the async operation, will be removed from the datamodel
 */
void usp_engine_process_async_command_success (struct usp_engine *engine, struct dm_object *reqobj) ARGS_NONNULL;


/**
 * @brief Process a failed async command execution and send notifications as required
 *
 * @param[in] engine	the engine used to process the command execution
 * @param[in] reqobj	the associated request object related to the async operation, will be removed from the datamodel
 * @param[in] err_code	the error code to use for the notification
 * @param[in] err_msg	the error message to use for the notification
 */
void usp_engine_process_async_command_error (struct usp_engine *engine, struct dm_object *reqobj, unsigned int err_code, char *err_msg) ARGS_NONNULL;


/**
 * @brief Return the number of in-flight messages between the USP Agent and a Controller
 *
 * @param[in] engine	the engine to get the communication count from
 *
 * @return the number of current messages in flight
 */
int usp_engine_get_communication_count (struct usp_engine *engine) ARGS_NONNULL;


/**
 * @brief Prepare a clean shutdown by giving all components a chance to clean up ongoing activities
 *
 * @param[in] engine	the engine to shut down
 */
void usp_engine_shutdown (struct usp_engine *engine);


/**
 * @brief Shuts down and deallocated a USP engine
 *
 * @param[in] engine	the engine to shut down and deallocate
 *
 * Note: This function will also destroy the event base!
 */
void usp_engine_free (struct usp_engine *engine) ARGS_NONNULL;

#endif /* USP_ENGINE_H_ */
