/**
 * @file
 *
 * @copyright Axiros GmbH 2019-2021, All rights reserved
 *
 * @brief USP MTP registry function definitions
 */


#include "compiler.h"

#ifndef USP_MTP_H_
#define USP_MTP_H_

/* Forward declaration */
struct evbuffer;
struct usp_engine;


/**
 *  @brief Our private usp_mtp structure
 */
struct usp_mtp;

enum usp_mtp_connection_state
{
	MTP_STATE_IDLE = 0,
	MTP_STATE_LISTENING,
	MTP_STATE_CONNECTED,
	MTP_STATE_BUSY,
	MTP_STATE_ERROR,
};

typedef enum usp_mtp_connection_state (*mtp_connect_cb) (struct usp_mtp *mtp);
typedef enum usp_mtp_connection_state (*mtp_disconnect_cb) (struct usp_mtp *mtp);
typedef int (*mtp_send_cb) (struct usp_mtp *mtp, char *recipient, struct evbuffer *data);
typedef enum usp_mtp_connection_state (*mtp_state_cb) (struct usp_mtp *mtp);
typedef void (*mtp_free_cb) (void *mtp);


/**
 * @brief Register a new MTP implementation
 *
 * @param[in] engine		the engine to register the MTP for
 * @param[in] send_cb		the callback to call to send out a USP record via this MTP
 * @param[in] state_cb		the callback to call to identify the current state of this MTP
 * @param[in] connect_cb	the callback to call to connect this MTP to a peer
 * @param[in] disconnect_cb	the callback to call to disconnect this MTP from a peer
 * @param[in] epid		the USP Controller Endpoint ID associated with this MTP
 * @param[in] priv		MTP implementation private data stored for later access
 *
 * @return a new USP MTP object
 */
struct usp_mtp *usp_mtp_register_mtp (struct usp_engine *engine, mtp_send_cb send_cb, mtp_state_cb state_cb, mtp_connect_cb connect_cb, mtp_disconnect_cb disconnect_cb, mtp_free_cb free_cb, char *epid, void *priv) ARGS_NONNULL MALLOC RETURNS_NONNULL;


/**
 * @brief Unregister and free a previously registered MTP implementation
 *
 * @param[in] engine		the engine to remove the MTP from
 * @param[in] mtp		The MTP to free
 */
void usp_mtp_unregister_mtp (struct usp_engine *engine, struct usp_mtp *mtp) ARGS_NONNULL;


/**
 * @brief Attempt to connect the MTP to the configured peer
 *
 * @param[in] mtp	the MTP to connect
 *
 * @return the connection state of this MTP
 */
enum usp_mtp_connection_state usp_mtp_connect (struct usp_mtp *mtp) ARGS_NONNULL;


/**
 * @brief Attempt to disconnect the MTP from its peer
 *
 * @param[in] mtp	the MTP to disconnect
 *
 * @return the connection state of this MTP
 */
enum usp_mtp_connection_state usp_mtp_disconnect (struct usp_mtp *mtp) ARGS_NONNULL;


/**
 * @brief Get the current connection state of the MTP
 *
 * @param[in] mtp	the MTP to query the state for
 *
 * @return the connection state of this MTP
 */
enum usp_mtp_connection_state usp_mtp_get_connection_state (struct usp_mtp *mtp) ARGS_NONNULL;


/**
 * @brief Send a record stored in an evbuffer to a recipient using this MTP
 *
 * @param[in] mtp	the MTP to send the record out on
 * @param[in] recipient	the recipient ID for this record
 * @param[in] buf	the buffer containing the record to be sent out
 *
 * @note Use of this function will drain the buffer.
 *
 * @return the number of bytes sent to the recipient
 */
int usp_mtp_send_evbuffer (struct usp_mtp *mtp, char *recipient, struct evbuffer *buf) ARGS_NONNULL;


/**
 * @brief Return the pointer to the private MTP implementation handle
 *
 * @param[in] mtp	the MTP get the private data handle for
 *
 * @return the stored pointer to the MTP implementation private data
 */
void *usp_mtp_get_private (struct usp_mtp *mtp) ARGS_NONNULL;

#endif /* USP_MTP_H_ */
