/**
 * @file
 *
 * @copyright Axiros GmbH 2015-2021, All rights reserved
 *
 * @brief Definition of datamodel parameter access and handling functions
 */

#ifndef DMPARAMETER_H
#define DMPARAMETER_H

#include <stddef.h>

#include "compiler.h"
#include "dmpermission.h"
#include "datamodel/DataModel.h"


/* Forward declarations */
struct dm_parameter;
struct dm_parameter_template;


enum tr069_notification_type {PASSIVE_NOTIFY = 1, ACTIVE_NOTIFY = 2, PASSIVE_LW_NOTIFY = 4, ACTIVE_LW_NOTIFY = 8, ANY_NOTIFY = 15};


/**
 * @brief Prototype of a validation function to use when a parameter with
 *        syntax validation type dm_syn_custom needs to be validated
 */
typedef unsigned int (*dm_parameter_validation_func) (struct dm_parameter *param, char *value) ARGS_NONNULL;


enum dm_parameter_compare_result
{
	DM_PC_INVALID = 0,	/**< provided value is invalid */
	DM_PC_EQUAL = 1,	/**< the Left Operand is equal to the right one */
	DM_PC_LO_LESS = 2,	/**< the Left Operand less than the right one */
	DM_PC_LO_GREATER = 4,	/**< the Left Operand grater than the right one */
	DM_PC_NOT_EQUAL = 8,		/**< the Left Operand is different than the right one */
	DM_PC_NOT_SUPPORTED = 16,	/**< parameter type is not supported */
};


/**
 * @brief Mark parameter as changed to generate notifications
 *
 * @param[in] par The parameter to mark as changes
 *
 * @note The changed flag indicates that metadata around this parameter has
 *       changed and needs to be synced to disk as well as pending
 *       notifications
 */
void dm_parameter_set_changed (struct dm_parameter *par) ARGS_NONNULL;


/**
 * @brief Clear the changed flag from the parameter
 *
 * @param[in] par The parameter from which to remove the changed state
 *
 * @note The changed flag indicates that metadata around this parameter has
 *       changed and needs to be synced to disk as well as pending
 *       notifications
 */
void dm_parameter_clear_changed_state (struct dm_parameter *par) ARGS_NONNULL;


/**
 * @brief Retrieves the current notification setting of the specified parameter
 *
 * @param[in] par The parameter for which to query the notification state for
 *
 * @return notification value (according to TR-069) or 0 for off
 */
unsigned int dm_parameter_get_notification (struct dm_parameter *par) ARGS_NONNULL;


/**
 * @brief Returns whether the value of the specified parameter has changed and needs syncing to disk
 *
 * @param[in] par The parameter for which to query the changed state for
 *
 * @return 1 if value has changed since last sync, 0 otherwise
 */
unsigned int dm_parameter_has_changed (struct dm_parameter *par) ARGS_NONNULL;


/**
 * @brief Returns whether the specified parameter has a notification of the given type enabled
 *
 * @param[in] par The parameter for which to check whether notification is enabled
 * @param[in] type A bitmask specifying which notification types to check for
 *
 * @return 1 if the parameter has notifications of the given type enable
 */
unsigned int dm_parameter_notification_enabled (struct dm_parameter *par, enum tr069_notification_type type) ARGS_NONNULL;


/**
 * @brief Returns the object associated with this parameter, e.g. the table belonging to a NumberOfEntries parameter
 *
 * @param[in] par The parameter for which to find the associated object for
 *
 * @return The associated object if it is specified and exists, NULL otherwise
 */
struct dm_object *dm_parameter_get_associated_object (struct dm_parameter *par) ARGS_NONNULL;


/**
 * @brief Retrieves the default active notification setting of the specified parameter
 *
 * @param[in] par The parameter for which to query the default active notification state for
 *
 * @return 1 if parameter has active notification enabled by default, 0 if not
 */
unsigned int dm_parameter_get_default_notification (struct dm_parameter *par) ARGS_NONNULL;


/**
 * @brief Sets the notification settings of the specified parameter to the
 *	  specified setting and enter into the according notification list
 *
 * @param[in] par The parameter for which to set the notification setting
 * @param[in] notify The notification value to set
 */
void dm_parameter_set_notification (struct dm_parameter *par, unsigned int notify) ARGS_NONNULL;


/**
 * @brief Retrieves the current access setting of the specified parameter
 *
 * @param[in] par The parameter for which to query the access state for
 *
 * @return access value (according to TR-069, only 1 for subscriber ATM) or 0 for off
 */
unsigned int dm_parameter_get_access (struct dm_parameter *par) ARGS_NONNULL;


/**
 * @brief Sets the access settings of the specified parameter to the
 *	  specified value
 *
 * @param[in] par		The parameter for which to set the access setting
 * @param[in] access_type	The access value to set
 */
void dm_parameter_set_access (struct dm_parameter *par, unsigned int access_type) ARGS_NONNULL;


/**
 * @brief Sets the flag for this parameter to indicate whether the parameter
 *        value should be serialised to disk (i.e. persist) or not
 *
 * @param[in] par	The parameter for which to set or clear the needs_backup flag
 * @param[in] flag	The new flag value, 1 to persist the parameter value, 0 otherwise
 */
void dm_parameter_set_needs_backup (struct dm_parameter *par, unsigned int flag) ARGS_NONNULL;


/**
 * @brief Returns whether the parameter value should be made persistent
 *
 * @param[in] par The parameter for which to query the backup requirement
 *
 * @return 1 if parameter value should be kept persistently, 0 if not
 */
unsigned int dm_parameter_needs_backup (struct dm_parameter *par) ARGS_NONNULL;


/**
 * @brief Retrieves the current value of the specified parameter
 *
 * @param[in] par The parameter for which to query the current value for
 *
 * @return Current value of the specified parameter or the default value
 */
char *dm_parameter_get_value (struct dm_parameter *par) ARGS_NONNULL RETURNS_NONNULL;


/**
 * @brief Retrieves the current internal value of the specified parameter
 *
 * @param[in] par The parameter for which to query the current internal value for
 *
 * @return Current internal value of the specified parameter or NULL if not set
 */
char *dm_parameter_get_value_internal (struct dm_parameter *par) ARGS_NONNULL;


/**
 * @brief Compare the value of two parameters
 *
 * @param[in] par	One of the parameters to compare the value to
 * @param[in] other	The other of the parameters to compare the value to
 *
 * @return enum dm_parameter_compare_result
 */
enum dm_parameter_compare_result dm_parameter_compare (struct dm_parameter *par, struct dm_parameter *other) ARGS_NONNULL;


/**
 * @brief Type aware parameter's value comparison to another value
 *
 * @param[in] par	The parameter to use for the value comparison
 * @param[in] value	The value string to compare to
 *
 * @return enum dm_parameter_compare_result
 */
enum dm_parameter_compare_result dm_parameter_compare_value (struct dm_parameter *par, char *value) ARGS_NONNULL;


/**
 * @brief Compare the value of a parameter to another value with the given length
 *
 * @param[in] par	The parameters to use for the value comparison
 * @param[in] value	The value string to compare to
 * @param[in] len	The length of the string to compare to
 *
 * @return enum dm_parameter_compare_result
 */
enum dm_parameter_compare_result dm_parameter_compare_value_len (struct dm_parameter *par, char *value, size_t len) ARGS_NONNULL;


/**
 * @brief Returns whether the parameter is declared as a command parameter
 *
 * @param[in] par The parameter for which to check whether its a command parameter
 *
 * @return 1 if parameter is command parameter, 0 if not
 *
 * @details command parameters (sometimes referred to as action parameters) are
 *          supposed to trigger a specific activity and be read back as the {{null}}
 *          value of the parameter type
 */
unsigned int dm_parameter_is_command (struct dm_parameter *par) ARGS_NONNULL;


/**
 * @brief Returns whether the parameter is declared as a forced inform
 *        parameter, i.e. should be included in every inform message to the ACS
 *
 * @param[in] par The parameter for which to check whether it's forced inform
 *
 * @return 1 if parameter is forced inform, 0 if not
 */
unsigned int dm_parameter_is_forcedinform (struct dm_parameter *par) ARGS_NONNULL;


/**
 * @brief Returns whether the parameter is declared hidden, i.e. should be
 *        empty when read by the ACS
 *
 * @param[in] par The parameter for which to check whether its hidden
 *
 * @return 1 if parameter is hidden, 0 if not
 */
unsigned int dm_parameter_is_hidden (struct dm_parameter *par) ARGS_NONNULL;


/**
 * @brief Returns whether the parameter is used as a key
 *
 * @param[in] par The parameter for which to check whether its a key
 *
 * @return 1 if parameter is a key, 0 if not
 */
unsigned int dm_parameter_is_key (struct dm_parameter *par) ARGS_NONNULL;


/**
 * @brief Returns whether the parameter contains a list value
 *
 * @param[in] par The parameter for which to check it contains a list value
 *
 * @return 1 if parameter is a type taking a list value, 0 if not
 */
unsigned int dm_parameter_is_list (struct dm_parameter *par) ARGS_NONNULL;


/**
 * @brief Returns whether the parameter is used as a partial key
 *
 * @param[in] par The parameter for which to check whether its a partial key
 *
 * @return 1 if parameter is a partial key, 0 if it is a full key
 */
unsigned int dm_parameter_is_partial_key (struct dm_parameter *par) ARGS_NONNULL;


/**
 * @brief Returns whether the parameter is declared to deny active notification
 *
 * @param[in] par The parameter for which to check whether active notification is allowed
 *
 * @return 1 if parameter will allow active notification, 0 if not
 */
unsigned int dm_parameter_is_nonotify (struct dm_parameter *par) ARGS_NONNULL;


/**
 * @brief Returns whether the parameter is writable or read-only
 *
 * @param[in] par The parameter for which to check whether it is writable
 *
 * @return 1 if parameter will allow modification, 0 if it is read-only
 */
unsigned int dm_parameter_is_writable (struct dm_parameter *par) ARGS_NONNULL;


/**
 * @brief Returns whether the parameter is a mandatory to use input argument for a USP command
 *
 * @param[in] par The parameter for which to check whether it is mandatory
 *
 * @return 1 if parameter is an USP command input argument which is required to be used
 */
unsigned int dm_parameter_is_mandatory (struct dm_parameter *par) ARGS_NONNULL;


/**
 * @brief Returns whether the parameter is declared as a reference parameter
 *
 * @param[in] par The parameter for which to check whether its a reference parameter
 *
 * @return 1 if parameter is reference parameter, 0 if not
 */
unsigned int dm_parameter_is_reference (struct dm_parameter *par) ARGS_NONNULL;


/**
 * @brief Return the current containing object of this parameter
 *
 * @param[in] par The parameter for which to get the containing object
 *
 * @return The object containing this parameter or NULL if freestanding
 */
struct dm_object *dm_parameter_get_object (struct dm_parameter *par) ARGS_NONNULL;


/**
 * @brief Set the containing object of the specified parameter
 *
 * @param[in] par The parameter for which to set the containing object
 * @param[in] obj The containing object for this parameter, can also be NULL
 */
void dm_parameter_set_object (struct dm_parameter *par, struct dm_object *obj) ARGINDEX_NONNULL(1);


/**
 * @brief Returns whether two parameters use the same template
 *
 * @param[in] par The first parameter for which to compare the template
 * @param[in] par2 The second parameter for which to compare the template
 *
 * @return 1 if the parameters use the same template, 0 otherwise
 */
unsigned int dm_parameter_has_same_template (struct dm_parameter *par, struct dm_parameter *par2) ARGS_NONNULL;


/**
 * @brief Verify whether the reference is valid
 *
 * @details weak reference: it does not necessarily reference an existing parameter or object
 *          strong reference: it always either references a valid parameter or object, or else is a null reference
 *
 * @param[in] par	pointer to parameter structure
 * @param[in] ref	reference to check
 *
 * @return 1 is reference is valid, 0 otherwise
 */
unsigned int dm_parameter_reference_is_valid (struct dm_parameter *par, char* ref) ARGS_NONNULL;


/**
 * @brief Removes invalid references from parameter
 *
 * @param[in] par	pointer to parameter structure
 */
void dm_parameter_clear_invalid_reference (struct dm_parameter *par) ARGS_NONNULL;


/**
 * @brief Verify whether the parameter is a key type with a value unique to the object instance it is part of
 *
 * @param[in] par	pointer to parameter structure
 * @param[in] value	the value to check for uniqueness properties
 * @param[in] len	the length of the value
 *
 * @return 0 if no conflicts are detected, != 0 if there're are keying conflicts
 */
unsigned int dm_parameter_check_key_conflict (struct dm_parameter *par, char *value, size_t len) ARGS_NONNULL;


/**
 * @brief Verify whether the value would be applicable for this parameter
 *
 * @param[in] param pointer to parameter structure
 * @param[in] value new value to verify for applicability
 *
 * @return 0 if successful, or TR-069 error code
 */
unsigned int dm_parameter_verify_value (struct dm_parameter* param, char* value) ARGS_NONNULL;


/**
 * @brief Verify whether a normalized version of the value would be applicable for this parameter
 *
 * @param[in] param pointer to parameter structure
 * @param[in] value new value to verify for applicability after normalization
 *
 * @return 0 if successful, or TR-069 error code
 */
unsigned int dm_parameter_verify_normalized_value (struct dm_parameter *param, char *value) ARGS_NONNULL;


/**
 * @brief Sets the specified value for the parameter running the full set of checks
 *
 * @param[in] param		pointer to parameter structure
 * @param[in] passedvalue	new value
 *
 * @return 0 if successful, or TR-069 error code
 */
unsigned int dm_parameter_set_value (struct dm_parameter *param, char *passedvalue) ARGS_NONNULL;


/**
 * @brief Retrieves the partial name of the specified parameter, i.e. without object path
 *
 * @param[in] par The parameter for which to query the name for
 *
 * @return Name of the specified parameter
 */
char *dm_parameter_get_name (struct dm_parameter *par) ARGS_NONNULL RETURNS_NONNULL;


/**
 * @brief Return the permission matrix of the parameter
 *
 * @param[in] par The parameter to return the permissions for
 *
 * @return permission matrix
 */
dm_permission_t dm_parameter_get_permissions (struct dm_parameter *par) ARGS_NONNULL;


/**
 * @brief Set the permission matrix of the parameter
 *
 * @param[in] par The parameter to set the permissions for
 * @param[in] permissions The permission matrix to set on the parameter
 */
void dm_parameter_set_permissions (struct dm_parameter *par, dm_permission_t permissions) ARGS_NONNULL;


/**
 * @brief Retrieves the type of the specified parameter
 *
 * @param[in] par The parameter for which to query the type for
 *
 * @return Type of the parameter
 */
dm_xsd_type dm_parameter_get_type (struct dm_parameter *par) ARGS_NONNULL;


/**
 * @brief Returns pointer to a static string with TR-069/SOAP type name of the parameter
 *
 * @param[in] par The parameter for which to return the type name for
 *
 * @return pointer to an static string containing the type name
 *
 * @note The returned does *NOT* contain the SOAP namespace for the type
 */
const char *dm_parameter_get_type_name (struct dm_parameter *par) ARGS_NONNULL RETURNS_NONNULL;


/**
 * @brief Frees memory allocated by data in parameter
 *
 * @param[in] par The parameter for which to free the associated data
 */
void dm_parameter_free_data (struct dm_parameter *par) ARGS_NONNULL;


/**
 * @brief Free allocated parameter
 *
 * @param[in] par A pointer to the previously allocated parameter storage to free
 */
void dm_parameter_free (struct dm_parameter *par) ARGS_NONNULL;


/**
 * @brief Allocate and initialise a parameter structure using the specified template
 *
 * @param[in] tmpl The parameter info to use for the new parameter
 *
 * @return A freshly allocated parameter using information from the template
 */
struct dm_parameter *dm_parameter_init (struct dm_parameter_template *tmpl) ARGS_NONNULL RETURNS_NONNULL MALLOC;


/**
 * @brief Clone metadata from one parameter to another
 *
 * @param[in] dest The preallocated parameter structure to clone to
 * @param[in] orig The parameter to clone from
 */
void dm_parameter_copy_template (struct dm_parameter *dest, struct dm_parameter *orig) ARGS_NONNULL;


/**
 * @brief Create a new parameter cloning the metadata from the provided parameter
 *
 * @param[in] orig The parameter to clone from
 *
 * @return A freshly allocated parameter which is a clone of the one provided
 */
struct dm_parameter *dm_parameter_clone (struct dm_parameter *orig) ARGS_NONNULL RETURNS_NONNULL MALLOC;


/**
 * @brief Append a new parameter to the end of the parameter chain
 *
 * @param[in] anypar Any parameter in the chain at which end the new parameter should be added to
 * @param[in] par The parameter to be added to the parameter chain
 */
void dm_parameter_append (struct dm_parameter *anypar, struct dm_parameter *par) ARGS_NONNULL;


/**
 * @brief Find a parameter relative to the given one
 *
 * @param[in] par Parameter to use as a starting point
 * @param[in] relpath The path to use for searching starting from the current parameter
 *
 * @return The parameter found by following the search path or NULL if not found
 *
 * @details The relpath can either be a simple parameter name in which case it
 *          will be search in the same object as the specified starting point
 *          or it can be a fully qualified path or if it starts with a "." the
 *          search will start at the current component object or root.
 *
 *          Please note that the search will always start from the current
 *          parameter so if it is a detached parameter nothing will be found.
 */
struct dm_parameter *dm_parameter_find_related_parameter (struct dm_parameter *par, char *relpath) ARGS_NONNULL;


/**
 * @brief Get the next parameter following the given one
 *
 * @param[in] par The parameter for which to return the following
 * @return The parameter that follows the current one or NULL if it doesn't exist
 */
struct dm_parameter *dm_parameter_get_next (struct dm_parameter *par) ARGS_NONNULL;


/**
 * @brief Extracts the instantiated path name of the parameter into the specified buffer starting at the specified object
 *
 * @param[in] par	The parameter to return the relative path name for
 * @param[in] from	The parent of the object to include in the path, for the datamodel root this would be NULL
 * @param[in,out] buf	Buffer to store path in
 * @param[in] buflen	Size of buffer
 *
 * @note When obj == from or from == NULL, the path will just contain the parameter name
 */
void dm_parameter_extract_relative_path (struct dm_parameter *par, struct dm_object *from, char *buf, size_t buflen) ARGINDEX_NONNULL(1) ARGINDEX_NONNULL(3);


/**
 * @brief Extracts the supported path name of the parameter in TR-106 notation into the specified buffer starting at the specified object
 *
 * @param[in] par	The parameter to return the relative path name for
 * @param[in] from	The parent of the object to include in the path, for the datamodel root this would be NULL
 * @param[in,out] buf	Buffer to store path in
 * @param[in] buflen	Size of buffer
 *
 * @note When obj == from or from == NULL, the path will just contain the parameter name
 */
void dm_parameter_extract_relative_dm_path (struct dm_parameter *par, struct dm_object *from, char *buf, size_t buflen) ARGINDEX_NONNULL(1) ARGINDEX_NONNULL(3);


/**
 * @brief Output all information about the particular parameter to the system log
 *
 * @param[in] par	The parameter to explain
 * @param[in] loglevel	The loglevel to use
 */
void dm_parameter_explain (struct dm_parameter *par, int loglevel) ARGS_NONNULL;

#endif /* DMPARAMETER_H */
