/**
 * @file
 *
 * @copyright Axiros GmbH 2013-2017, All rights reserved
 *
 * @brief Cryptography support functionality
 */

#ifndef OSADAPTER_CRYPTO_H
#define OSADAPTER_CRYPTO_H

#include "compiler.h"

struct evbuffer;

/** @defgroup crypto Cryptographic functions
 * These functions are wrappers and general initialization or cleanup functions
 * for all crypto functionality. The default implementation integrates with
 * OpenSSL and/or mbedTLS.
 *
 * @{
 */

enum CRYPTO_HASH_TYPE
{
	CRYPTO_HASH_SHA1,
	CRYPTO_HASH_SHA224,
	CRYPTO_HASH_SHA256,
	CRYPTO_HASH_SHA384,
	CRYPTO_HASH_SHA512,
};


/**
 * @brief Initialize the crypto implementation
 *
 * @note This function contains guards to be called several times for on-demand initialization
 */
void crypto_init (void);


/**
 * @brief Returns whether crypto support is available or not
 *
 * @return 1 for available crypto support, 0 for none
 */
int crypto_is_available (void);


/**
 * @brief Set up a fresh SSL/TLS context for permanent use
 *
 * @param[in] ca_path		the full filename path of a CA certificate bundle
 * @param[in] client_cert	the full filename path of a client certificate, may be NULL if not applicable
 * @param[in] client_key	the full filename path of a client key to decrypt a certificate, may be NULL if not applicable
 */
void crypto_connection_create_TLS_context (char *ca_path, char *client_cert, char *client_key) ARGINDEX_NONNULL(1);


/**
 * @brief Free a previously set up SSL/TLS context
 */
void crypto_connection_destroy_TLS_context (void);


/**
 * @brief Setup the context for a TLS session with the specified host on the specified socket
 *
 * @param[in] fd		The fd descriptor of the open socket
 * @param[in] host		The name of the host we're connecting to (for Server Name Indication)
 * @param[in] ca_path		the full filename path of a CA certificate bundle
 * @param[in] client_cert	the full filename path of a client certificate, may be NULL if not applicable
 * @param[in] client_key 	the full filename path of a client key to decrypt a certificate, may be NULL if not applicable
 *
 * @return The implementation specific TLS session context or NULL if setup failed
 *
 * @note The setup includes the required handshake with the remote so the
 * socket needs to be working properly and without interference from other code
 */
void *crypto_setup_tls_connection (int fd, char *host, char *ca_path, char *client_cert, char *client_key) ARGINDEX_NONNULL(2) ARGINDEX_NONNULL(3);


/**
 * @brief Establish the encrypted connection
 *
 * @param connctx The context for the conntection to establish
 *
 * @return 1 if the encryption was successfully established, 0 if not
 */
int crypto_establish (void *connctx) ARGS_NONNULL;


/**
 * @brief Verify whether the given string is a usable PEM encoded certificate
 *
 * @param pem The string which is supposed to represent a PEM encoded certificate
 *
 * @return 1 if the string was identified as PEM encoded certificate, 0 otherwise
 */
int crypto_check_is_certificate (char *pem) ARGS_NONNULL;


/**
 * @brief Return the issuer of the PEM encoded certificate
 *
 * @param pem The string which is supposed to represent a PEM encoded certificate
 *
 * @return an allocated string containing the issuer or NULL if certificate is invalid
 */
char *crypto_certificate_get_issuer (char *pem) ARGS_NONNULL;


/**
 * @brief Return the serial number of the PEM encoded certificate
 *
 * @param pem The string which is supposed to represent a PEM encoded certificate
 *
 * @return an allocated string containing the serialnumber or NULL if certificate is invalid
 */
char *crypto_certificate_get_serialnumber (char *pem) ARGS_NONNULL;


/**
 * @brief Return the fingerprint of the PEM encoded certificate
 *
 * @param typ The hash mechanism to use to generate the fingerprint
 * @param pem The string which is supposed to represent a PEM encoded certificate
 *
 * @return an allocated string containing the fingerprint or NULL if certificate is invalid or hash algorithm not implemented
 */
char *crypto_certificate_get_fingerprint (enum CRYPTO_HASH_TYPE typ, char *pem) ARGS_NONNULL;


/**
 * @brief Return whether TLS connection is properly established
 *
 * @param connctx The connection context we would like to check
 *
 * @return 1 if connection is fully established, 0 if not
 */
int crypto_is_established (void *connctx) ARGS_NONNULL;


/**
 * @brief Check for a correctly established TLS connection
 *
 * @param host The name of the host we're connecting to CN and host name validation
 * @param connctx A pointer to implementation specific TLS connection context
 * @param abstime 1 if the system has absolute time currently, 0 if not
 *
 * @return 1 for a correctly established connection,
 */
int crypto_verify_tls_connection (char *host, void *connctx, int abstime) ARGINDEX_NONNULL (1);


/**
 * @brief Shut down an established TLS connection
 *
 * @param connctx A void pointer to TLS context to shut down
 */
void crypto_close_tls_connection (void *connctx) ARGS_NONNULL;


/**
 * @brief Free internal structures associated with this connection context
 *
 * @param connctx A void pointer containing the TLS context to be freed
 *
 * @note An implementation of this function MUST NOT affect global state
 */
void crypto_free_connection_context (void *connctx) ARGS_NONNULL;


/**
 * @brief Decrypt data from the established connection using the provided
 *        context and put unencrypted data into provided buffer
 *
 * @param connctx A pointer to implementation specific TLS connection context
 * @param buf A pointer to an evbuffer which will be filled with unencrypted data
 *
 * @return The amount of unencrypted bytes, 0 for connection closed or -1 for error
 */
int crypto_decrypt_data (void *connctx, struct evbuffer *buf) ARGS_NONNULL;


/**
 * @brief Encrypt data from the provided buffer (draining it) and send it
 *        encrypted over the established connection using the provided
 *
 * @param connctx A pointer to implementation specific TLS connection context
 * @param buf A pointer to an evbuffer containing the unencrypted data
 *
 * @return The amount of bytes drained from the provided buffer or -1 for error
 */
int crypto_encrypt_data (void *connctx, struct evbuffer *buf) ARGS_NONNULL;


/**
 * @brief Clean up the used crypto implementation
 */
void crypto_cleanup (void);

/**@}*/

#endif /* OSADAPTER_CRYPTO_H */
