/**
 * @file
 *
 * @copyright Axiros GmbH 2010-2017, All rights reserved
 *
 * @brief OS abstraction for system functionality
 */

#ifndef OSADAPT_SYSTEM_H_
#define OSADAPT_SYSTEM_H_

#include <time.h>

#include "config.h"
#include "compiler.h"

#include "network.h"

struct tr069_engine;
struct frontend;


 /*! \page ossys System Interface Functions
The \ref ossysf are called from AXACT to interact with the device. They have to be adapted depending on the device
HW functions and provided interfaces.

\section sosys1 Basic functions:

The basic functions are
\li \ref ossys_reboot_device Reboot of the CPE
\li \ref ossys_reset_device Factory reset of the CPE
\li \ref ossys_get_uptime of the CPE

\section sosys2 Custom Hook:
The following functions provide a hook to add and remove custom functions, which execute tasks in the background,
e.g. libevent timer to check network status, or query internal interfaces.
\li \ref ossys_backgroundcheck_init Initialize the background task(s) called during AXACT initialization.
\li \ref ossys_backgroundcheck_cleanup Cleanup the background task(s) called during AXACT shutdown.

\section sosys4 Connection request url update mechanism
These functions provide a mechanism to keep the connection request URL current.
\li \ref ossys_crurl_update_trigger

\section sosys3 Pipe Handling
These functions connect and disconnect the named pipe which is used from AXACT for command interface. Usually
they can be used unchanged.
\li \ref ossys_connect_pipe
\li \ref ossys_disconnect_pipe

Back to \ref interfaces "Interfaces".
*/

/** @defgroup ossysf System Interface Functions
 * These are the main functions called from AXACT to interact with the device. They have to be adapted depending on the device HW and
interface provided.
 * @{
 */

/**
 * @brief Read the current uptime in s
 *
 * @return uptime in seconds
 */
time_t ossys_get_uptime (void);


/**
 * @brief The SOAP error to return when the ACS wants to issue a Reboot RPC
 *
 * This function should check whether the ACS may issue a Reboot command to
 * execute after the ongoing session. If the value 0 is returned, success will
 * be indicated to the ACS and a restart of the system queued and executed
 * via ossys_reboot_device() (see below) after the ongoing session. Otherwise
 * the error code will be indicated to the ACS, permissible standard errors are
 * 9001, 9002, 9003.
 *
 * @param fe	The frontend to use
 *
 * @return the status code to return to the ACS
 */
unsigned int ossys_allow_reboot (struct frontend *fe) ARGS_NONNULL;


/**
 * @brief Reboot the CPE
 *
 * This function should reboot the device
 *
 * @note no cleanup needed, the function will be called after
 * the client core has been shut down
 *
 * @param fe	The frontend to use
 */
void ossys_reboot_device (struct frontend *fe) ARGS_NONNULL;


/**
 * @brief The SOAP error to return when the ACS wants to issue a FactoryReset RPC
 *
 * This function should check whether the ACS may issue a FactoryReset command
 * to execute after the ongoing session. If the value 0 is returned, success
 * will be indicated to the ACS and a restart of the system queued and executed
 * via ossys_reset_device() (see below) after the ongoing session. Otherwise
 * the error code will be indicated to the ACS, permissible standard errors are
 * 9000, 9001, 9002, 9003.
 *
 * @param fe	The frontend to use
 *
 * @return the status code to return to the ACS
 */
unsigned int ossys_allow_reset (struct frontend *fe) ARGS_NONNULL;

/**
 * @brief Reset the CPE to factory defaults
 *
 * @param fe		The frontend to use
 * @param[in] origin	indicates the originator of the requested reset:
 *   				  0: reset was invoked from the command line
 *   				  1: reset was invoked from the acs,
 *   				  2: reset was invoked internally (due to error)
 *
 * This function should reset the device to its factory defaults
 *
 * @note no cleanup needed, the function will be  called after
 * the client core has been shut down
 */
void ossys_reset_device (struct frontend *fe, int origin) ARGS_NONNULL;


/**
 * @brief Create a named pipe and connect to it
 *
 * @param[in] path the full path of the fifo that should be created, the last
 *	      directory of which needs to exist
 *
 * @return fd if successful, -1 on error
 *
 * @details this function will create the a fifo at the given path and connect
 * 	    a socket to it returning the fd for communication
 */
int ossys_connect_pipe (char *path) ARGS_NONNULL;


/**
 * @brief Disconnect and remove a named pipe
 *
 * @param[in] path the full path of the fifo that has been created previously
 * @param[in] fd to the created socket
 *
 * @details this function will disconnect the socket and remote the fifo at the
 *          given path
 */
void ossys_disconnect_pipe (char *path, int fd) ARGS_NONNULL;


/**
 * @brief Callback trigger for update of connection request url
 *
 * @param[in] fe	the frontend to use for data updates
 * @param[in] callback	the function to execute after the new IP was determined
 */
void ossys_crurl_update_trigger (struct frontend *fe, network_probe_cb callback) ARGS_NONNULL;


/**
 * @brief Initialize mechanism to do system background checks
 *
 * @param fe	The frontend to use
 *
 * @details Use this function to initialize any system background checks that
 *          should be done on the system. This can be used to set up a timer to
 *          monitor crucial system parameters or other mechanism to do this.
 *
 */
void ossys_backgroundcheck_init (struct frontend *fe) ARGS_NONNULL;


/**
 * @brief Cleanup mechanism for system background checks
 *
 * @param fe	The frontend to use
 *
 */
void ossys_backgroundcheck_cleanup (struct frontend *fe) ARGS_NONNULL;


/**
 * @brief Get the first use date
 *
 * @param fe	The frontend to use
 *
 * @return The time which is supposed to be the time of first use
 */
time_t ossys_get_first_use_date (struct frontend *fe) ARGS_NONNULL;


#if defined(ENABLE_TR069)
/**
 * @brief Checks whether this device has to sent a bootstrap or not
 *
 * @param fe	The frontend to use
 *
 * @return 1 if we currently BOOTSTRAPping, 0 if this is a consecutive run
 */
int ossys_is_bootstrap (struct frontend *fe) ARGS_NONNULL;


/**
 * @brief Sets a flag to indicate that we already did a bootstrap
 *
 * @param fe	The frontend to use
 *
 */
void ossys_set_bootstrap (struct frontend *fe) ARGS_NONNULL;


/**
 * @brief Resets the flag that indicates that we already did a bootstrap
 *
 * @param fe	The frontend to use
 *
 */
void ossys_clear_bootstrap (struct frontend *fe) ARGS_NONNULL;


/**
 * @brief Starts a new status write transaction
 *
 * @param fe	The frontend to use
 *
 * @return 1 if we could successfully open up a new transaction
 */
int ossys_begin_status (struct frontend *fe) ARGS_NONNULL;

/**
 * @brief Callback function to be used to deserialise single TR-069 events
 *
 * @param event Event code of the read event
 * @param when Execution time of event
 * @param cmdkey Commandkey of event
 * @param details Additional event details
 * @param engine The engine in which this event was created on
 *
 * @return 0 for failure, 1 for success
**/
typedef int (*ossys_parse_event)(const char *event, char *cmdkey, time_t when, char *details, struct tr069_engine *engine) ARGS_NONNULL;


/**
 * @brief Read events from the status storage and call specified function for each event
 *
 * @param fe		The frontend to use
 * @param parse_cb	Callback function to call for each individual element
 *
 * @return 1 if successful, 0 if not
 */
int ossys_read_status (struct frontend *fe, ossys_parse_event parse_cb) ARGS_NONNULL;


/**
 * @brief Serialise a single event into the status storage
 *
 * @param fe		The frontend to use
 * @param event		Name of the event to serialise (i.e. TR-069 event code)
 * @param cmdkey	Eventcode of the event, may be NULL if none
 * @param when		Start time of event, maybe 0 if ASAP
 * @param details	Additional details to store for later use, may be NULL
 *
 * @return 1 if successful, 0 if not
 */
int ossys_write_status (struct frontend *fe, const char *event, char *cmdkey, time_t when, char *details) ARGINDEX_NONNULL (1);


/**
 * @brief End the currently open status transaction
 *
 * @param fe	The frontend to use
 *
 * @return 1 if we could successfully close the current status transaction
 */
int ossys_end_status (struct frontend *fe) ARGS_NONNULL;


/**
 * @brief Clear all serialised status items
 *
 * @param fe	The frontend to use
 *
 * @return 1 if we could successfully clear all status items
 */
int ossys_clear_status (struct frontend *fe) ARGS_NONNULL;
#endif

/**@}*/

#endif /* OSADAPT_SYSTEM_H_ */
